from conan.tools.gnu.pkgconfigdeps.pc_files_templates import get_alias_pc_filename_and_content, \
    get_pc_filename_and_content
from conan.tools.gnu.pkgconfigdeps.pc_info_loader import get_package_with_components_info, \
    get_single_package_info, get_aliases_info


def _get_aliases_pc_files_and_content(conanfile, dep, aliases_info, root_pc_files):
    """
    Get all the PC files and content for the aliases defined previously
    for package and components names
    """
    pc_files = {}
    for name, aliases in aliases_info.items():
        for alias in aliases:
            pc_name, pc_content = get_alias_pc_filename_and_content(
                dep,
                alias,
                [name],  # require is the own name which is being used the aliases for.
                description="Alias %s for %s" % (alias, name),
            )
            if pc_name in pc_files:
                conanfile.output.warn("[%s] The PC alias name %s already exists and it matches with "
                                      "another alias one. Please, review all the "
                                      "pkg_config_aliases defined. Skipping it!"
                                      % (dep, pc_name))
            elif pc_name in root_pc_files:
                conanfile.output.warn("[%s] The PC alias name %s already exists and it matches with "
                                      "another package/component one. Please, review all the "
                                      "pkg_config_aliases defined. Skipping it!"
                                      % (dep, pc_name))
            else:
                pc_files[pc_name] = pc_content
    return pc_files


def _get_components_pc_files_and_content(conanfile, dep, package_info, components_info):
    """
    Get the PC files and content for dependency's components
    """
    pc_files = {}
    # Loop through all the package's components
    for pc_info_component in components_info:
        # Get the *.pc file content for each component
        description = "Conan component: %s-%s" % (package_info.name, pc_info_component.name)
        pc_name, pc_content = get_pc_filename_and_content(
            conanfile,
            dep,
            pc_info_component.name,
            pc_info_component.requires,
            description,
            cpp_info=pc_info_component.cpp_info)
        if pc_name in pc_files:
            conanfile.output.warn("[%s] The PC component name %s already exists and it matches with "
                                  "another component one. Please, review all the "
                                  "component's pkg_config_name defined. Skipping it!"
                                  % (dep, pc_name))
        else:
            pc_files[pc_name] = pc_content
    return pc_files


def _get_package_with_components_pc_files_and_content(conanfile, dep, package_info, components_info):
    """
    Get the PC files and content for dependencies with components.
    The PC files will be saved in this order:
        1- Package components.
        2- Root component.

    Note: If the root-package PC name matches with any other of the components one, the first one
    is not going to be created. Components have more priority than root package.
    """
    pc_files = {}
    # First, let's load all the components PC files
    pc_files.update(_get_components_pc_files_and_content(conanfile, dep, package_info, components_info))
    description = "Conan package: %s" % package_info.name
    pc_name, pc_content = get_alias_pc_filename_and_content(
        dep,
        package_info.name,
        package_info.requires,
        description
    )
    # Second, let's load the root package's PC file ONLY
    # if it does not already exist in components one
    # Issue related: https://github.com/conan-io/conan/issues/10341
    if pc_name in pc_files:
        conanfile.output.warn("[%s] The PC package name %s already exists and it matches with "
                              "another component one. Please, review all the "
                              "component's pkg_config_name defined. Skipping it!" % (dep, pc_name))
    else:
        pc_files[pc_name] = pc_content
    return pc_files


def _get_single_package_pc_file_and_content(conanfile, dep, package_info):
    """
    Get the PC files for dependencies without components
    """
    description = "Conan package: %s" % package_info.name
    pc_name, pc_content = get_pc_filename_and_content(conanfile, dep, package_info.name,
                                                      package_info.requires, description)
    return {pc_name: pc_content}


def get_pc_files_and_content(conanfile, dep):
    """
    Get all the PC files given a dependency (package, components and alias ones)
    """
    pc_files = {}
    if dep.cpp_info.has_components:  # Package with components
        pkg_info, components_info = get_package_with_components_info(dep)
        pc_files.update(_get_package_with_components_pc_files_and_content(conanfile, dep, pkg_info,
                                                                          components_info))
    else:
        # Package without components
        pkg_info = get_single_package_info(dep)
        components_info = []
        pc_files.update(_get_single_package_pc_file_and_content(conanfile, dep, pkg_info))
    # Package and components names aliases
    aliases_info = get_aliases_info(dep, pkg_info, components_info)
    pc_files.update(_get_aliases_pc_files_and_content(conanfile, dep, aliases_info, pc_files))
    return pc_files



from collections import namedtuple

from conans.errors import ConanException


_PCInfoComponent = namedtuple("PCInfoComponent", ['name', 'requires', 'cpp_info', 'ref_name'])
_PCInfoPackage = namedtuple('PCInfoPackage', ['name', 'requires'])


def _get_name_with_namespace(namespace, name):
    """Build a name with a namespace, e.g., openssl-crypto"""
    return "%s-%s" % (namespace, name)


def _get_package_reference_name(dep):
    """Get the reference name for the given package"""
    # FIXME: this str(dep.ref.name) is only needed for python2.7 (unicode values).
    #        Remove it for Conan 2.0
    return str(dep.ref.name)


def _get_package_aliases(dep):
    pkg_aliases = dep.cpp_info.get_property("pkg_config_aliases")
    return pkg_aliases or []


def _get_component_aliases(dep, comp_name):
    if comp_name not in dep.cpp_info.components:
        # foo::foo might be referencing the root cppinfo
        if dep.ref.name == comp_name:
            return _get_package_aliases(dep)
        raise ConanException("Component '{name}::{cname}' not found in '{name}' "
                             "package requirement".format(name=_get_package_reference_name(dep),
                                                          cname=comp_name))
    comp_aliases = dep.cpp_info.components[comp_name].get_property("pkg_config_aliases")
    return comp_aliases or []


def _get_package_name(dep):
    pkg_name = dep.cpp_info.get_property("pkg_config_name")
    return pkg_name or _get_package_reference_name(dep)


def _get_component_name(dep, comp_name):
    if comp_name not in dep.cpp_info.components:
        # foo::foo might be referencing the root cppinfo
        if dep.ref.name == comp_name:
            return _get_package_name(dep)
        raise ConanException("Component '{name}::{cname}' not found in '{name}' "
                             "package requirement".format(name=_get_package_reference_name(dep),
                                                          cname=comp_name))
    comp_name = dep.cpp_info.components[comp_name].get_property("pkg_config_name")
    return comp_name


def _get_component_requires_names(dep, cpp_info):
    """
    Get all the pkg-config valid names from the requires ones given a dependency and
    a CppInfo object.

    Note: CppInfo could be coming from one Component object instead of the dependency
    """
    dep_ref_name = _get_package_reference_name(dep)
    ret = []
    for req in cpp_info.requires:
        pkg_ref_name, comp_ref_name = req.split("::") if "::" in req \
                                                      else (dep_ref_name, req)
        if dep_ref_name != pkg_ref_name:
            req_conanfile = dep.dependencies.host[pkg_ref_name]
        else:
            req_conanfile = dep
        comp_name = _get_component_name(req_conanfile, comp_ref_name)
        if not comp_name:
            pkg_name = _get_package_name(req_conanfile)
            comp_name = _get_name_with_namespace(pkg_name, comp_ref_name)
        ret.append(comp_name)
    return ret


def get_single_package_info(dep):
    """
    Get the whole package information when it does not have components declared
    """
    pkg_name = _get_package_name(dep)
    # At first, let's check if we have defined some component requires, e.g., "pkg::cmp1"
    requires = _get_component_requires_names(dep, dep.cpp_info)
    # If we have found some component requires it would be enough
    if not requires:
        # If no requires were found, let's try to get all the direct dependencies,
        # e.g., requires = "other_pkg/1.0"
        requires = [_get_package_name(req) for req in dep.dependencies.direct_host.values()]
    # Save the package information
    pkg_info = _PCInfoPackage(pkg_name, requires)
    return pkg_info


def get_package_with_components_info(dep):
    """
    Get the whole package and its components information like their own requires, names and even
    the cpp_info for each component.
    """
    pkg_name = _get_package_name(dep)
    pkg_requires = []
    components_info = []
    # Loop through all the package's components
    for comp_name, comp_cpp_info in dep.cpp_info.get_sorted_components().items():
        comp_requires_names = _get_component_requires_names(dep, comp_cpp_info)
        pkg_comp_name = _get_component_name(dep, comp_name)
        if not pkg_comp_name:
            pkg_comp_name = _get_name_with_namespace(pkg_name, comp_name)
        pkg_requires.append(pkg_comp_name)
        # Save each component information
        comp_info = _PCInfoComponent(pkg_comp_name, comp_requires_names, comp_cpp_info, comp_name)
        components_info.append(comp_info)
    # Save the package information
    pkg_info = _PCInfoPackage(pkg_name, pkg_requires)
    return pkg_info, components_info


def get_aliases_info(dep, package_info, components_info):
    """
    Get the whole aliases information for the given package and components names already calculated
    """
    aliases_info = {}
    # Package aliases
    if package_info:
        pkg_aliases = _get_package_aliases(dep)
        if pkg_aliases:
            aliases_info[package_info.name] = pkg_aliases
    # Component aliases
    for comp_info in components_info:
        comp_aliases = _get_component_aliases(dep, comp_info.ref_name)
        if comp_aliases:
            aliases_info[comp_info.name] = comp_aliases
    return aliases_info



import os
import textwrap

from jinja2 import Template, StrictUndefined

from conan.tools.gnu.gnudeps_flags import GnuDepsFlags


def _get_pc_file_template():
    return textwrap.dedent("""\
        {%- macro get_libs(libdirs, cpp_info, gnudeps_flags) -%}
        {%- for _ in libdirs -%}
        {{ '-L"${libdir%s}"' % loop.index + " " }}
        {%- endfor -%}
        {%- for sys_lib in (cpp_info.libs + cpp_info.system_libs) -%}
        {{ "-l%s" % sys_lib + " " }}
        {%- endfor -%}
        {%- for shared_flag in (cpp_info.sharedlinkflags + cpp_info.exelinkflags) -%}
        {{  shared_flag + " " }}
        {%- endfor -%}
        {%- for framework in (gnudeps_flags.frameworks + gnudeps_flags.framework_paths) -%}
        {{ framework + " " }}
        {%- endfor -%}
        {%- endmacro -%}

        {%- macro get_cflags(includedirs, cxxflags, cflags, defines) -%}
        {%- for _ in includedirs -%}
        {{ '-I"${includedir%s}"' % loop.index + " " }}
        {%- endfor -%}
        {%- for cxxflag in cxxflags -%}
        {{ cxxflag + " " }}
        {%- endfor -%}
        {%- for cflag in cflags-%}
        {{ cflag + " " }}
        {%- endfor -%}
        {%- for define in defines-%}
        {{  "-D%s" % define + " " }}
        {%- endfor -%}
        {%- endmacro -%}

        prefix={{ prefix_path }}
        {% for path in libdirs %}
        {{ "libdir{}={}".format(loop.index, path) }}
        {% endfor %}
        {% for path in includedirs %}
        {{ "includedir%d=%s" % (loop.index, path) }}
        {% endfor %}
        {% if pkg_config_custom_content %}
        # Custom PC content
        {{ pkg_config_custom_content }}
        {% endif %}

        Name: {{ name }}
        Description: {{ description }}
        Version: {{ version }}
        Libs: {{ get_libs(libdirs, cpp_info, gnudeps_flags) }}
        Cflags: {{ get_cflags(includedirs, cxxflags, cflags, defines) }}
        {% if requires|length %}
        Requires: {{ requires|join(' ') }}
        {% endif %}
    """)


def _get_alias_pc_file_template():
    return textwrap.dedent("""\
        Name: {{ name }}
        Description: {{ description }}
        Version: {{ version }}
        {% if requires|length %}
        Requires: {{ requires|join(' ') }}
        {% endif %}
    """)


def _get_formatted_dirs(folders, prefix_path_):
    ret = []
    for i, directory in enumerate(folders):
        directory = os.path.normpath(directory).replace("\\", "/")
        prefix = ""
        if not os.path.isabs(directory):
            prefix = "${prefix}/"
        elif directory.startswith(prefix_path_):
            prefix = "${prefix}/"
            directory = os.path.relpath(directory, prefix_path_).replace("\\", "/")
        ret.append("%s%s" % (prefix, directory))
    return ret


def get_pc_filename_and_content(conanfile, dep, name, requires, description, cpp_info=None):
    package_folder = dep.package_folder or ""  # If editable, package_folder can be None
    cpp_info = cpp_info or dep.cpp_info
    version = cpp_info.get_property("component_version") or dep.ref.version

    prefix_path = package_folder.replace("\\", "/")
    libdirs = _get_formatted_dirs(cpp_info.libdirs, prefix_path)
    includedirs = _get_formatted_dirs(cpp_info.includedirs, prefix_path)
    custom_content = cpp_info.get_property("pkg_config_custom_content")

    context = {
        "prefix_path": prefix_path,
        "libdirs": libdirs,
        "includedirs": includedirs,
        "pkg_config_custom_content": custom_content,
        "name": name,
        "description": description,
        "version": version,
        "requires": requires,
        "cpp_info": cpp_info,
        "cxxflags": [var.replace('"', '\\"') for var in cpp_info.cxxflags],
        "cflags": [var.replace('"', '\\"') for var in cpp_info.cflags],
        "defines": [var.replace('"', '\\"') for var in cpp_info.defines],
        "gnudeps_flags": GnuDepsFlags(conanfile, cpp_info)
    }
    template = Template(_get_pc_file_template(), trim_blocks=True, lstrip_blocks=True,
                        undefined=StrictUndefined)
    return name + ".pc", template.render(context)


def get_alias_pc_filename_and_content(dep, name, requires, description):
    context = {
        "name": name,
        "description": description,
        "version": dep.ref.version,
        "requires": requires
    }
    template = Template(_get_alias_pc_file_template(), trim_blocks=True,
                        lstrip_blocks=True, undefined=StrictUndefined)
    return name + ".pc", template.render(context)



import json
import os
import platform

from conan.tools.cmake.layout import get_build_folder_custom_vars
from conan.tools.cmake.utils import is_multi_configuration
from conan.tools.microsoft import is_msvc
from conans.errors import ConanException
from conans.util.files import save, load


def _build_preset(conanfile, multiconfig):
    build_type = conanfile.settings.get_safe("build_type")
    configure_preset_name = _configure_preset_name(conanfile, multiconfig)
    ret = {"name": _build_preset_name(conanfile),
           "configurePreset": configure_preset_name}
    if multiconfig:
        ret["configuration"] = build_type
    return ret


def _build_preset_name(conanfile):
    build_type = conanfile.settings.get_safe("build_type")
    custom_conf = get_build_folder_custom_vars(conanfile)
    if custom_conf:
        if build_type:
            return "{}-{}".format(custom_conf, build_type.lower())
        else:
            return custom_conf
    return build_type.lower() if build_type else "default"


def _configure_preset_name(conanfile, multiconfig):
    build_type = conanfile.settings.get_safe("build_type")
    custom_conf = get_build_folder_custom_vars(conanfile)

    if multiconfig or not build_type:
        return "default" if not custom_conf else custom_conf

    if custom_conf:
        return "{}-{}".format(custom_conf, str(build_type).lower())
    else:
        return str(build_type).lower()


def _configure_preset(conanfile, generator, cache_variables, toolchain_file, multiconfig):
    build_type = conanfile.settings.get_safe("build_type")
    name = _configure_preset_name(conanfile, multiconfig)
    if not multiconfig and build_type:
        cache_variables["CMAKE_BUILD_TYPE"] = build_type
    ret = {
            "name": name,
            "displayName": "'{}' config".format(name),
            "description": "'{}' configure using '{}' generator".format(name, generator),
            "generator": generator,
            "cacheVariables": cache_variables,
           }
    if "Ninja" in generator and is_msvc(conanfile):
        toolset_arch = conanfile.conf.get("tools.cmake.cmaketoolchain:toolset_arch")
        if toolset_arch:
            toolset_arch = "host={}".format(toolset_arch)
            ret["toolset"] = {
                "value": toolset_arch,
                "strategy": "external"
            }
        arch = {"x86": "x86",
                "x86_64": "x64",
                "armv7": "ARM",
                "armv8": "ARM64"}.get(conanfile.settings.get_safe("arch"))

        if arch:
            ret["architecture"] = {
                "value": arch,
                "strategy": "external"
            }

    if not _forced_schema_2(conanfile):
        ret["toolchainFile"] = toolchain_file
    else:
        ret["cacheVariables"]["CMAKE_TOOLCHAIN_FILE"] = toolchain_file

    if conanfile.build_folder:
        # If we are installing a ref: "conan install <ref>", we don't have build_folder, because
        # we don't even have a conanfile with a `layout()` to determine the build folder.
        # If we install a local conanfile: "conan install ." with a layout(), it will be available.
        ret["binaryDir"] = conanfile.build_folder
    return ret


def _forced_schema_2(conanfile):
    version = conanfile.conf.get("tools.cmake.cmaketoolchain.presets:max_schema_version",
                                 check_type=int)
    if not version:
        return False

    if version < 2:
        raise ConanException("The minimum value for 'tools.cmake.cmaketoolchain.presets:"
                             "schema_version' is 2")
    if version < 4:
        return True

    return False


def _schema_version(conanfile, default):
    if _forced_schema_2(conanfile):
        return 2

    return default


def _contents(conanfile, toolchain_file, cache_variables, generator):
    """
    Contents for the CMakePresets.json
    It uses schema version 3 unless it is forced to 2
    """
    ret = {"version": _schema_version(conanfile, default=3),
           "cmakeMinimumRequired": {"major": 3, "minor": 15, "patch": 0},
           "configurePresets": [],
           "buildPresets": [],
           "testPresets": []
          }
    multiconfig = is_multi_configuration(generator)
    ret["buildPresets"].append(_build_preset(conanfile, multiconfig))
    _conf = _configure_preset(conanfile, generator, cache_variables, toolchain_file, multiconfig)
    ret["configurePresets"].append(_conf)
    return ret


def write_cmake_presets(conanfile, toolchain_file, generator, cache_variables):
    cache_variables = cache_variables or {}
    if platform.system() == "Windows" and generator == "MinGW Makefiles":
        if "CMAKE_SH" not in cache_variables:
            cache_variables["CMAKE_SH"] = "CMAKE_SH-NOTFOUND"

        cmake_make_program = conanfile.conf.get("tools.gnu:make_program", default=cache_variables.get("CMAKE_MAKE_PROGRAM"))
        if cmake_make_program:
            cmake_make_program = cmake_make_program.replace("\\", "/")
            cache_variables["CMAKE_MAKE_PROGRAM"] = cmake_make_program

    if "CMAKE_POLICY_DEFAULT_CMP0091" not in cache_variables:
        cache_variables["CMAKE_POLICY_DEFAULT_CMP0091"] = "NEW"

    preset_path = os.path.join(conanfile.generators_folder, "CMakePresets.json")
    multiconfig = is_multi_configuration(generator)

    if os.path.exists(preset_path):
        data = json.loads(load(preset_path))
        build_preset = _build_preset(conanfile, multiconfig)
        position = _get_already_existing_preset_index(build_preset["name"], data["buildPresets"])
        if position is not None:
            data["buildPresets"][position] = build_preset
        else:
            data["buildPresets"].append(build_preset)

        configure_preset = _configure_preset(conanfile, generator, cache_variables, toolchain_file,
                                             multiconfig)
        position = _get_already_existing_preset_index(configure_preset["name"],
                                                      data["configurePresets"])
        if position is not None:
            data["configurePresets"][position] = configure_preset
        else:
            data["configurePresets"].append(configure_preset)
    else:
        data = _contents(conanfile, toolchain_file, cache_variables, generator)

    data = json.dumps(data, indent=4)
    save(preset_path, data)
    save_cmake_user_presets(conanfile, preset_path)


def save_cmake_user_presets(conanfile, preset_path):
    # Try to save the CMakeUserPresets.json if layout declared and CMakeLists.txt found
    if conanfile.source_folder and conanfile.source_folder != conanfile.generators_folder:
        if os.path.exists(os.path.join(conanfile.source_folder, "CMakeLists.txt")):
            """
                Contents for the CMakeUserPresets.json
                It uses schema version 4 unless it is forced to 2
            """
            user_presets_path = os.path.join(conanfile.source_folder, "CMakeUserPresets.json")
            if not os.path.exists(user_presets_path):
                data = {"version": _schema_version(conanfile, default=4),
                        "vendor": {"conan": dict()}}
            else:
                data = json.loads(load(user_presets_path))
                if "conan" not in data.get("vendor", {}):
                    # The file is not ours, we cannot overwrite it
                    return
            data = _append_user_preset_path(conanfile, data, preset_path)
            data = json.dumps(data, indent=4)
            save(user_presets_path, data)


def _get_already_existing_preset_index(name, presets):
    """Get the index of a Preset with a given name, this is used to replace it with updated contents
    """
    positions = [index for index, p in enumerate(presets)
                 if p["name"] == name]
    if positions:
        return positions[0]
    return None


def _append_user_preset_path(conanfile, data, preset_path):
    """ - Appends a 'include' to preset_path if the schema supports it.
        - Otherwise it merges to "data" all the configurePresets, buildPresets etc from the
          read preset_path.
    """
    if not _forced_schema_2(conanfile):
        if "include" not in data:
            data["include"] = []
        # Clear the folders that have been deleted
        data["include"] = [i for i in data.get("include", []) if os.path.exists(i)]
        if preset_path not in data["include"]:
            data["include"].append(preset_path)
        return data
    else:
        # Merge the presets
        cmake_preset = json.loads(load(preset_path))
        for preset_type in ("configurePresets", "buildPresets", "testPresets"):
            for preset in cmake_preset.get(preset_type, []):
                if preset_type not in data:
                    data[preset_type] = []

                position = _get_already_existing_preset_index(preset["name"], data[preset_type])
                if position is not None:
                    # If the preset already existed, replace the element with the new one
                    data[preset_type][position] = preset
                else:
                    data[preset_type].append(preset)
        return data


def load_cmake_presets(folder):
    tmp = load(os.path.join(folder, "CMakePresets.json"))
    return json.loads(tmp)


def get_configure_preset(cmake_presets, conanfile):
    expected_name = _configure_preset_name(conanfile, multiconfig=False)
    # Do we find a preset for the current configuration?
    for preset in cmake_presets["configurePresets"]:
        if preset["name"] == expected_name:
            return preset

    expected_name = _configure_preset_name(conanfile, multiconfig=True)
    # In case of multi-config generator or None build_type
    for preset in cmake_presets["configurePresets"]:
        if preset["name"] == expected_name:
            return preset

    # FIXME: Might be an issue if someone perform several conan install that involves different
    #        CMake generators (multi and single config). Would be impossible to determine which
    #        is the correct configurePreset because the generator IS in the configure preset.

    raise ConanException("Not available configurePreset, expected name is {}".format(expected_name))
